"""Module for kind of useful things that don't really belong anywhere else (just yet)."""

from .activation import *
from .data_helpers import *
from .data_structures import *
from .heaviside import *
