"""Module for algorithms that pre-process the data in some way."""

from .adjust_labels import *
from .biased_split import *
from .domain_adaptation import *
from .feature_binning import *
from .scaling import *
from .splits import *
